#include "fields.h"
#include <time.h>
#include <stdio.h>
#include <curses.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/pstat.h>
#include <signal.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/dk.h>
#include "sorting.h"
#include <ctype.h>

static int getwinwidth(WINDOW *w) {
	int x1,x2;
	int y1,y2;
	getbegyx(w,y1,x1);
	getmaxyx(w,y2,x2);
	return x2-x1;
}
static int getwinheight(WINDOW *w) {
	int x1,x2;
	int y1,y2;
	getbegyx(w,y1,x1);
	getmaxyx(w,y2,x2);
	return y2-y1;
}

static const char *lookup_username(long userid) {
	static long last_userid=-1234;
	static char buf[64];
	if(userid==last_userid) return buf;
	struct passwd *pw = getpwuid((uid_t)userid);
	if(!pw)
		return "?";
	strncpy(buf,pw->pw_name,sizeof(buf));
	buf[sizeof(buf)-1] = '\0';
	last_userid=userid;
	return buf;
}

#define NO_USER -123
static int lookup_userid(const char *username) {
	struct passwd *pw = getpwnam(username);
	if(!pw)
		return NO_USER;
	return pw->pw_uid;
}


static FieldDefinition field_def[]={
	{	5,5,5,		"Pid"		},
	{	4,8,8,		"User"		},
	{	3,3,3,		"Pri"		},
	{	3,3,3,		"NI"		},
	{	4,9,5,		"vsz"		},
	{	4,9,5,		"res"		},
	{	4,9,5,		"Share"		},
	{	4,4,4,		"Stat"		},
	{	3,6,6,		"%cpu"		},
	{	3,6,6,		"%mem"		},
	{	5,5,5,		"time"		},
	{	8,128,16,	"Command"	},
};
static const unsigned fields = 12;

static time_t grab_time;
static time_t uptime;
static double loadavg[3];

static unsigned proccount_total;
static unsigned proccount_sleep;
static unsigned proccount_running;
static unsigned proccount_stopped;
static unsigned proccount_zombie;
static unsigned proccount_idle;
static unsigned proccount_other;

struct StateTime {
	long usr;
	long nice;
	long sys;
	long idle;
	long intr;
	double usr_pct;
	double nice_pct;
	double sys_pct;
	double idle_pct;
	double intr_pct;
};

unsigned cpus;

struct CPU {
	unsigned id;
	bool online;
	StateTime state_time;
};
static CPU *cpu=0;
static StateTime global_state_time;

static unsigned mem_total_kb;
static unsigned mem_used_kb;
static unsigned mem_free_kb;
static unsigned swap_total_kb;
static unsigned swap_used_kb;
static unsigned swap_free_kb;

struct Process {
	pid_t		pid;
	uid_t		userid;
	int         priority;
	int         nice;
	unsigned	vsz_kb;
	unsigned	res_kb;
	unsigned	share_kb;
	unsigned	state;
	double		cpu_usage;
	double		mem_usage;
	bool		is_idle;
	time_t		start_time;
	time_t		cpu_time;
	char		ucomm[PST_UCOMMLEN+1];
	char        cmd[PST_CLEN+1];
};
static Process *process=0;
static unsigned processes=0;
static unsigned process_alloc=0;

#define wrapchange(x,y) ((x)<(y)?((y)-(x)):((x)-(y)))
static void update_state_time(StateTime *state_time, const _T_LONG_T cpu_time[]) {
	StateTime new_state_time;
	new_state_time.usr = cpu_time[CP_USER];
	new_state_time.nice = cpu_time[CP_NICE];
	new_state_time.sys = cpu_time[CP_SYS] + cpu_time[CP_WAIT] + cpu_time[CP_BLOCK] + cpu_time[CP_SWAIT] + cpu_time[CP_SSYS];
	new_state_time.idle = cpu_time[CP_IDLE];
	new_state_time.intr = cpu_time[CP_INTR];
	
	long prev_total = state_time->usr+state_time->nice+state_time->sys+state_time->idle+state_time->intr;
	long new_total = new_state_time.usr+new_state_time.nice+new_state_time.sys+new_state_time.idle+new_state_time.intr;
	
	long total_change = wrapchange(prev_total,new_total);
	if(total_change==0) total_change=1;
	
	long l;
	l = wrapchange(state_time->usr,new_state_time.usr);
	new_state_time.usr_pct = (l * 100.0) / total_change;
	l = wrapchange(state_time->nice,new_state_time.nice);
	new_state_time.nice_pct = (l * 100.0) / total_change;
	l = wrapchange(state_time->sys,new_state_time.sys);
	new_state_time.sys_pct = (l * 100.0) / total_change;
	l = wrapchange(state_time->idle,new_state_time.idle);
	new_state_time.idle_pct = (l * 100.0) / total_change;
	l = wrapchange(state_time->intr,new_state_time.intr);
	new_state_time.intr_pct = (l * 100.0) / total_change;
	
	*state_time = new_state_time;
}

static void tab2spc(char *s) {
    while(*s) {
        if(*s<' ') *s=' ';
        s++;
    }
}

void grab_info(bool filter_idle, const char *filter_user) {
	struct pst_static s_pst_static;
	memset(&s_pst_static,sizeof(s_pst_static),0);
	pstat_getstatic(&s_pst_static, sizeof(s_pst_static),1,0);
	struct pst_dynamic s_pst_dynamic;
	memset(&s_pst_dynamic,sizeof(s_pst_dynamic),0);
	pstat_getdynamic(&s_pst_dynamic,sizeof(s_pst_dynamic),1,0);
	struct pst_vminfo s_pst_vminfo;
	pstat_getvminfo(&s_pst_vminfo,sizeof(s_pst_vminfo),1,0);
	
	
	grab_time = time(0);
	uptime = grab_time - s_pst_static.boot_time;

	loadavg[0] = s_pst_dynamic.psd_avg_1_min;
	loadavg[1] = s_pst_dynamic.psd_avg_5_min;
	loadavg[2] = s_pst_dynamic.psd_avg_15_min;
	
	update_state_time(&global_state_time,s_pst_dynamic.psd_cpu_time);
	
	if(cpu) { free(cpu); cpu=0; }
	cpus = (unsigned)s_pst_dynamic.psd_max_proc_cnt;
	cpu = (CPU*)malloc(sizeof(CPU)*cpus);
	for(int i=0; i<cpus; i++) {
		struct pst_processor s_pst_processor;
		pstat_getprocessor(&s_pst_processor,sizeof(s_pst_processor),1,i);
		cpu[i].id = (int)s_pst_processor.psp_logical_id;
		cpu[i].online = s_pst_processor.psp_processor_state==PSP_SPU_ENABLED;
		update_state_time(&cpu[i].state_time,s_pst_processor.psp_cpu_time);
	}
	
	long page_size = s_pst_static.page_size;
	unsigned page_size_kb = (unsigned)(page_size/1024);
	
	//memory
	mem_total_kb = (unsigned)(s_pst_static.physical_memory * page_size_kb);
	mem_used_kb = (unsigned)(s_pst_dynamic.psd_rm * page_size_kb);
	mem_free_kb = (unsigned)(s_pst_dynamic.psd_free * page_size_kb);
	swap_total_kb = (unsigned)(s_pst_vminfo.psv_swapspc_max * page_size_kb);
	swap_used_kb = (unsigned)(s_pst_vminfo.psv_swapspc_cnt * page_size_kb);
	swap_free_kb = swap_total_kb-swap_used_kb;
	//Note: HP-UX 11.23 on itanium seems to return ...ahem.. weird values
	
	if(process) { free(process); process=0; }
	process = (Process*)malloc(sizeof(Process)*128);
	process_alloc=128;
	processes = 0;
	
	if(filter_user && !filter_user[0]) filter_user=0;
	uid_t filter_userid;
	if(filter_user) {
		filter_userid = lookup_userid(filter_user);
	}
    
    int proc_index=0;
    for(;;) {
       	struct pst_status pst_status_buf[10];
        int procs_returned = pstat_getproc(pst_status_buf,sizeof(pst_status_buf[0]),10,proc_index);
        if(procs_returned<=0) break;
        
        for(int i=0; i<procs_returned; i++) {
		struct pst_status *s = pst_status_buf+i;
    		if(filter_user && s->pst_uid!=filter_userid) continue;
		if(filter_idle && s->pst_stat==PS_IDLE) continue;
		if(processes>=process_alloc) {
			process = (Process*)realloc(process,sizeof(Process)*(process_alloc+128));
			process_alloc+=128;
		}
		//pstat_getproc frequently returns duplicates
		for(int j=0; j<processes; j++) {
			if(process[j].pid == s->pst_pid) goto next_process;
		}
		Process *p;
		p = process+processes;
		p->pid = (pid_t)s->pst_pid;
		p->userid = (uid_t)s->pst_uid;
		p->priority = (int)s->pst_pri;
		p->nice = (int)s->pst_nice;
		p->vsz_kb = (unsigned)(s->pst_vtsize+s->pst_vdsize+s->pst_vshmsize+s->pst_vmmsize)*page_size_kb;
		p->res_kb = (unsigned)(s->pst_rssize)*page_size_kb;
		p->share_kb = (unsigned)(s->pst_vshmsize)*page_size_kb;
p->share_kb += (unsigned)(s->pst_vmmsize)*page_size_kb;
		p->state = (unsigned)s->pst_stat;
		p->cpu_usage = s->pst_pctcpu*100.0;
		p->mem_usage = (p->res_kb*100.0)/mem_total_kb;
		p->start_time = s->pst_start;
		p->cpu_time = s->pst_utime + s->pst_stime;
		strncpy(p->ucomm,s->pst_ucomm,sizeof(p->ucomm));
		p->ucomm[sizeof(p->ucomm)-1]='\0';
		tab2spc(p->ucomm);
		strncpy(p->cmd,s->pst_cmd,sizeof(p->cmd));
		p->cmd[sizeof(p->cmd)-1]='\0';
		tab2spc(p->cmd);
		p->is_idle = p->state==PS_IDLE || p->cpu_usage<=0.02;
		processes++;
		next_process:;
	}
        proc_index += procs_returned;
    }
    
    proccount_total = processes;
    proccount_sleep = 0;
    proccount_running = 0;
    proccount_stopped = 0;
    proccount_zombie = 0;
    proccount_idle = 0;
    proccount_other = 0;
    
    
    
    for(unsigned i=0; i<processes; i++) {
        switch(process[i].state) {
            case PS_SLEEP:  proccount_sleep++; break;
            case PS_RUN:    proccount_running++; break;
            case PS_STOP:   proccount_stopped++; break;
            case PS_ZOMBIE: proccount_zombie++; break;
            case PS_IDLE:   proccount_idle++; break;
            case PS_OTHER:  proccount_other++; break;
        }
    }
}

void paint_top(WINDOW *w) {
	wclear(w);
	
	struct tm tm_grabtime = *localtime(&grab_time);
	char grab_time_buf[128];
	strftime(grab_time_buf,sizeof(grab_time_buf),"%H:%M",&tm_grabtime);
	char uptime_buf[128];
	time_t uptime_days = uptime/86400;
	time_t uptime_hours = (uptime-uptime_days*86400)/3600;
	time_t uptime_minutes = (uptime-uptime_days*86400-uptime_hours*3600)/60;
	if(uptime_days<1)
		sprintf(uptime_buf,"%d:%02d",(int)uptime_hours,(int)uptime_minutes);
	else
		sprintf(uptime_buf,"%d days %02d:%02d",(int)uptime_days,(int)uptime_hours,(int)uptime_minutes);
	
	wmove(w,0,0);
	wprintw(w,(char*)"%s  up %s,  load average: %4.2f, %4.2f, %4.2f",
	        grab_time_buf,
	        uptime_buf,
	        loadavg[0],
	        loadavg[1],
	        loadavg[2]
	       );
	
	wmove(w,1,0);
	wprintw(w,(char*)"%u processes: %u sleeping, %u running, %u stopped, %u zombie, %u idle, %u other",
	        proccount_total,proccount_sleep,proccount_running,proccount_stopped,proccount_zombie,proccount_idle,proccount_other
	       );
}

void paint_cpu(WINDOW *w, bool total_only)
{
	wclear(w);
	if(total_only) {
		wmove(w,0,0);
		wprintw(w,(char*)"CPU states:  %5.2f%% user, %5.2f%% nice, %5.2f%% sys, %5.2f%% idle, %5.2f%% intr",
		        global_state_time.usr_pct,
		        global_state_time.nice_pct,
		        global_state_time.sys_pct,
		        global_state_time.idle_pct,
		        global_state_time.intr_pct
		       );
	} else {
		for(unsigned i=0; i<cpus; i++) {
			if(i==0) {
				wmove(w,0,0);
				wprintw(w,(char*)"CPU states: ");
			} else
				wmove(w,i,12);
			wprintw(w,(char*)"%3u", cpu[i].id);
			if(cpu[i].online)
				wprintw(w,(char*)" %5.2f%% user, %5.2f%% nice, %5.2f%% sys, %5.2f%% idle, %5.2f%% intr",
		                cpu[i].state_time.usr_pct,
		                cpu[i].state_time.nice_pct,
		                cpu[i].state_time.sys_pct,
		                cpu[i].state_time.idle_pct,
		                cpu[i].state_time.intr_pct
		       );
		}
	}
}

void paint_memory(WINDOW *w) {
	wclear(w);
	wmove(w,0,0);
	wprintw(w,(char*)"Mem:  %7uK total, %7uK used, %7uK free",
	        mem_total_kb,
		    mem_used_kb,
		    mem_free_kb
	       );
	wmove(w,1,0);
	wprintw(w,(char*)"Swap: %7uK total, %7uK used, %7uK free",
		    swap_total_kb,
		    swap_used_kb,
		    swap_free_kb
		   );
}

void paint_header(WINDOW *w) {
	int width[32];
	DistributeFields(field_def,fields,getwinwidth(w),width);
	PaintHeaderFields(w,field_def,fields,width);
}

static void paint_kb(WINDOW *w, unsigned kb, int width) {
	if((kb<=9999999 && width>=7) ||
       (kb<= 999999 && width>=6) ||
	   (kb<=99999              )
      )
		wprintw(w,(char*)"%*u",width,kb);
	else
		wprintw(w,(char*)"%*uM",width-1,kb/1024);
}

static void paint_pct(WINDOW *w, int width, double pct) {
	if(width<5)
		wprintw(w,(char*)"%*.0f",width,pct);
	else
		wprintw(w,(char*)"%*.2f",width,pct);
}

static const char *state_text(unsigned state) {
    switch(state) {
        case PS_SLEEP:  return "s";
        case PS_RUN:    return "r";
        case PS_STOP:   return "stop";
        case PS_ZOMBIE: return "Z";
        case PS_IDLE:   return "I";
        case PS_OTHER:  return "?";
    }
	return "?";
}

#ifdef __cplusplus
extern "C" {
#endif
static int cmp_pid(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	return p1->pid - p2->pid;
}
static int cmp_age(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	return (int)(p1->start_time - p2->start_time);
}
static int cmp_cpuusage(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	if(p1->cpu_usage > p2->cpu_usage) return -1;
	if(p1->cpu_usage < p2->cpu_usage) return  1;
	return 0;
}
static int cmp_resmem(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	if(p1->res_kb > p2->res_kb) return -1;
	if(p1->res_kb < p2->res_kb) return  1;
	return 0;
}
static int cmp_cputime(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	if(p1->cpu_time > p2->cpu_time) return -1;
	if(p1->cpu_time < p2->cpu_time) return  1;
	return 0;
}
#ifdef __cplusplus
}
#endif

static void sort_processes(Sorting_t sorting) {
	switch(sorting) {
		case sorting_pid:      qsort(process,(size_t)processes,sizeof(process[0]), cmp_pid);      break;
		case sorting_age:      qsort(process,(size_t)processes,sizeof(process[0]), cmp_age);      break;
		case sorting_cpuusage: qsort(process,(size_t)processes,sizeof(process[0]), cmp_cpuusage); break;
		case sorting_resmem:   qsort(process,(size_t)processes,sizeof(process[0]), cmp_resmem);   break;
		case sorting_cputime:  qsort(process,(size_t)processes,sizeof(process[0]), cmp_cputime);  break;
	}
}

void paint_processes(WINDOW *w, bool filter_idle, const char *filter_user, Sorting_t sorting) {

	sort_processes(sorting);
	
	if(filter_user && !filter_user[0]) filter_user=0;
	uid_t filter_userid;
	if(filter_user) {
		filter_userid = lookup_userid(filter_user);
	}
	
	int width[32];
	DistributeFields(field_def,fields,getwinwidth(w),width);
	unsigned height = getwinheight(w);
	
	wclear(w);
	int line=0;
	for(unsigned i=0; i<processes; i++) {
		if(line>=height) break;
		Process *p = process+i;
		if(filter_idle && p->is_idle) continue;
		if(filter_user && p->userid!=filter_userid) continue;
		
		const int *wi=width;
		wmove(w,line,0);
		line++;
		
		wprintw(w,(char*)"%*lu",*wi,(unsigned long)p->pid);
		waddch(w,' ');
		wi++;
		
		const char *username=lookup_username((long)p->userid);
		wprintw(w,(char*)"%-*.*s",*wi,*wi,username);
		waddch(w,' ');
		wi++;
		
		wprintw(w,(char*)"%*d",*wi,p->priority);
		waddch(w,' ');
		wi++;
		
		wprintw(w,(char*)"%*d",*wi,p->nice);
		waddch(w,' ');
		wi++;
		
		paint_kb(w,p->vsz_kb,*wi);
		waddch(w,' ');
		wi++;
		
		paint_kb(w,p->res_kb,*wi);
		waddch(w,' ');
		wi++;
		
		paint_kb(w,p->share_kb,*wi);
		waddch(w,' ');
		wi++;
		
		wprintw(w,(char*)"%-*.*s",*wi,*wi,state_text(p->state));
		waddch(w,' ');
		wi++;
		
		paint_pct(w,*wi,p->cpu_usage);
		waddch(w,' ');
		wi++;
		
		paint_pct(w,*wi,p->mem_usage);
		waddch(w,' ');
		wi++;
		
		time_t cputime_minutes = p->cpu_time/60;
		time_t cputime_seconds = p->cpu_time%60;
		char cputime_str[16];
		sprintf(cputime_str,(char*)"%ld:%02ld", (long)cputime_minutes, (long)cputime_seconds);
		wprintw(w,(char*)"%*.*s",*wi,*wi,cputime_str);
		waddch(w,' ');
		wi++;
		
		if(strlen(p->ucomm)<=*wi)
			wprintw(w,(char*)"%-*.*s",*wi,*wi,p->ucomm);
		else {
			//fit as much of the command as possible
			const char *s=p->cmd;
			while(*s=='/' || isalnum(*s)) s++;
			/*const char *pend=s;*/
			while(s>p->cmd && *s!='/') s--;
			if(s>p->cmd) s++;
			while(strlen(s)<*wi-3 && s>p->cmd) s--;
			wprintw(w,(char*)"%-*.*s",*wi,*wi,s);
		}
	}
}


void do_kill(pid_t pid, int sig, WINDOW *err_window) {
	if(kill(pid,sig)!=0) {
		int save_errno=errno;
		wclear(err_window);
		wprintw(err_window,(char*)"signal(%ld,%d) failed, errno=%d, error=%s",(long)pid,sig,save_errno,strerror(save_errno));
		wgetch(err_window);
	}
}

void do_renice(pid_t pid, int value, WINDOW *err_window) {
	errno=0;
	if(setpriority(PRIO_PROCESS,pid,value)==-1 && errno!=0) {
		int save_errno=errno;
		wclear(err_window);
		wprintw(err_window,(char*)"setpriority(...,%ld,%d) failed, errno=%d, error=%s",(long)pid,value,save_errno,strerror(save_errno));
		wgetch(err_window);
	}
}
